@@ -9,8 +9,12 @@ See add_append_functionality_to_store_cls docs for examples.
 
 import time
 import types
+from typing import Callable, Optional
 
 from dol.trans import store_decorator
+from dol.util import exhaust
+
+utc_now = time.time
 
 
 def define_extend_as_seq_of_appends(obj):
@@ -172,7 +176,7 @@ class mk_item2kv_for:
         return item2kv
 
     @staticmethod
-    def utc_key(offset_s=0.0):
+    def utc_key(offset_s=0, factor=1, *, time_postproc: Optional[Callable] = None):
         """Make an item2kv function that uses the current time as the key, and the unchanged item as a value.
         The offset_s, which is added to the output key, can be used, for example, to align to another system's clock,
         or to get a more accurate timestamp of an event.
@@ -198,15 +202,19 @@ class mk_item2kv_for:
         >>> assert v == 'some data'  # just the item itself
 
         """
-        if offset_s == 0.0:  # splitting for extra speed (important in real time apps)
+        if time_postproc is None:
 
             def item2kv(item):
-                return time.time(), item
+                # Note: If real time accuracy is needed, you should use your own optimized
+                # item2kv function.
+                return factor * utc_now() + offset_s, item
 
         else:
 
             def item2kv(item):
-                return time.time() + offset_s, item
+                # Note: If real time accuracy is needed, you should use your own optimized
+                # item2kv function.
+                return time_postproc(factor * utc_now() + offset_s), item
 
         return item2kv
 
@@ -303,9 +311,7 @@ class mk_item2kv_for:
 
 
 @store_decorator
-def appendable(
-    store_cls=None, *, item2kv,
-):
+def appendable(store_cls=None, *, item2kv, return_keys=False):
     """Makes a new class with append (and consequential extend) methods
 
     Args:
@@ -347,15 +353,31 @@ def appendable(
     >>> s = appendable(d, item2kv=item_to_kv)
     >>> s.append({'L': 'let', 'I': 'it', 'G': 'go'}); list(s.items())
     [(('go', 'let'), {'I': 'it'})]
+
+    You can make the "append" and "extend" methods to return the new generated keys by
+    using the "return_keys" flag.
+
+    >>> d = {}
+    >>> s = appendable(d, item2kv=item_to_kv, return_keys=True)
+    >>> s.append({'L': 'let', 'I': 'it', 'G': 'go'})
+    ('go', 'let')
     """
 
     def append(self, item):
         k, v = item2kv(item)
         self[k] = v
+        if return_keys:
+            return k
 
     def extend(self, items):
-        for item in items:
-            self.append(item)
+        def gen_keys():
+            for item in items:
+                yield self.append(item)
+
+        gen = gen_keys()
+        if return_keys:
+            return list(gen)
+        exhaust(gen)
 
     return type(
         'Appendable' + store_cls.__name__,
