@@ -37,6 +37,7 @@ from collections.abc import (
 from typing import Any, Iterable, Tuple, Callable, Union, Optional
 
 from dol.util import wraps, _disabled_clear_method
+from dol.signatures import Sig
 
 Key = Any
 Val = Any
@@ -73,6 +74,8 @@ class AttrNames:
     KvPersister = (MutableMapping | {'head'}) - {'__reversed__'} - {'clear'}
 
 
+# TODO: Consider using ContainmentChecker and Sizer attributes which dunders would
+#  point to.
 class Collection(CollectionABC):
     """The same as collections.abc.Collection, with some modifications:
     - Addition of a ``head``
@@ -85,7 +88,7 @@ class Collection(CollectionABC):
         Therefore it may not be efficient, and in most cases, a method specific to the case should be used.
         :return: True if k is in the collection, and False if not
         """
-        for existing_x in self.__iter__():
+        for existing_x in iter(self):
             if existing_x == x:
                 return True
         return False
@@ -99,7 +102,7 @@ class Collection(CollectionABC):
         """
         # Note: Found that sum(1 for _ in self.__iter__()) was slower for small, slightly faster for big inputs.
         count = 0
-        for _ in self.__iter__():
+        for _ in iter(self):
             count += 1
         return count
 
@@ -128,50 +131,75 @@ class Collection(CollectionABC):
 #         return False
 
 
-class KvReader(Collection, Mapping):
+class MappingViewMixin:
+    KeysView: type = BaseKeysView
+    ValuesView: type = BaseValuesView
+    ItemsView: type = BaseItemsView
+
+    def keys(self) -> KeysView:
+        return self.KeysView(self)
+
+    def values(self) -> ValuesView:
+        return self.ValuesView(self)
+
+    def items(self) -> ItemsView:
+        return self.ItemsView(self)
+
+
+class KvReader(MappingViewMixin, Collection, Mapping):
     """Acts as a Mapping abc, but with default __len__ (implemented by counting keys)
     and head method to get the first (k, v) item of the store"""
 
     def head(self):
+        """Get the first (key, value) pair"""
         for k, v in self.items():
             return k, v
 
     def __reversed__(self):
         """The __reversed__ is disabled at the base, but can be re-defined in subclasses.
-        Rationale: KvReader is meant to wrap a variety of storage backends or key-value perspectives thereof.
-        Not all of these would have a natural or intuitive order nor do we want to maintain one systematically.
+        Rationale: KvReader is meant to wrap a variety of storage backends or key-value
+        perspectives thereof.
+        Not all of these would have a natural or intuitive order nor do we want to
+        incur the cost of maintaining one systematically.
 
         If you need a reversed list, here's one way to do it, but note that it
-        depends on how self iterates, which is not even assured to be consistent at every call:
-        ```
-        reversed = list(self)[::-1]
-        ```
+        depends on how self iterates, which is not even assured to be consistent at
+        every call:
+
+        .. code-block:: python
+
+            reversed = list(self)[::-1]
+        
 
         If the keys are comparable, therefore sortable, another natural option would be:
-        ```
-        reversed = sorted(self)[::-1]
-        ```
+        
+        .. code-block:: python
+
+            reversed = sorted(self)[::-1]
+        
         """
         raise NotImplementedError(__doc__)
 
 
-Reader = KvReader  # alias
+Reader = KvReader  # alias for back-compatibility
 
 
 # TODO: Should we really be using MutableMapping if we're disabling so many of it's methods?
 # TODO: Wishful thinking: Define store type so the type is defined by it's methods, not by subclassing.
 class KvPersister(KvReader, MutableMapping):
-    """ Acts as a MutableMapping abc, but disabling the clear and __reversed__ method,
+    """Acts as a MutableMapping abc, but disabling the clear and __reversed__ method,
     and computing __len__ by iterating over all keys, and counting them.
 
     Note that KvPersister is a MutableMapping, and as such, is dict-like.
     But that doesn't mean it's a dict.
 
     For instance, consider the following code:
-    ```
+
+    .. code-block:: python
+
         s = SomeKvPersister()
         s['a']['b'] = 3
-    ```
+    
     If `s` is a dict, this would have the effect of adding a ('b', 3) item under 'a'.
     But in the general case, this might
     - fail, because the `s['a']` doesn't support sub-scripting (doesn't have a `__getitem__`)
@@ -229,6 +257,8 @@ class DelegatedAttribute:
 
     def __get__(self, instance, owner):
         if instance is None:
+            # return getattr(getattr(owner, self.delegate_name), self.attr_name)
+            # return getattr(owner, self.attr_name, None)
             return self  # .__wrapped__  # TODO: The __wrapped__ would make it hard to debug
         else:
             # i.e. return instance.delegate.attr
@@ -328,8 +358,8 @@ def delegator_wrap(
     In this case, ``delegator_wrap`` returns a callable (class or function) that has the
     same signature as obj, but that produces instances that are wrapped by ``delegator``
 
-    :param delegator: An instance wrapper. A Callable (type or function -- with only one required input)
-        that will return a wrapped version of it's input instance.
+    :param delegator: An instance wrapper. A Callable (type or function -- with only
+        one required input) that will return a wrapped version of it's input instance.
     :param obj: The object (class or instance) to be wrapped.
     :return: A wrapped object
 
@@ -398,7 +428,12 @@ def delegator_wrap(
             type_decorator = delegate_to(
                 obj, class_trans=class_trans, delegation_attr=delegation_attr
             )
-            return type_decorator(delegator)
+            wrap = type_decorator(delegator)
+            try:  # try to give the wrap the signature of obj (if it has one)
+                wrap.__signature__ = Sig(obj)
+            except ValueError:
+                pass
+            return wrap
 
         else:
             assert isinstance(delegator, Callable)
@@ -424,18 +459,19 @@ class Store(KvPersister):
         __iter__    calls:	            _key_of_id
 
 
-    >>> # Default store: no key or value conversion ################################################
+    >>> # Default store: no key or value conversion #####################################
+    >>> from dol import Store
     >>> s = Store()
     >>> s['foo'] = 33
     >>> s['bar'] = 65
     >>> assert list(s.items()) == [('foo', 33), ('bar', 65)]
     >>> assert list(s.store.items()) == [('foo', 33), ('bar', 65)]  # see that the store contains the same thing
     >>>
-    >>> ################################################################################################
-    >>> # Now let's make stores that have a key and value conversion layer #############################
-    >>> # input keys will be upper cased, and output keys lower cased ##################################
-    >>> # input values (assumed int) will be converted to ascii string, and visa versa #################
-    >>> ################################################################################################
+    >>> #################################################################################
+    >>> # Now let's make stores that have a key and value conversion layer ##############
+    >>> # input keys will be upper cased, and output keys lower cased ###################
+    >>> # input values (assumed int) will be converted to ascii string, and visa versa ##
+    >>> #################################################################################
     >>>
     >>> def test_store(s):
     ...     s['foo'] = 33  # write 33 to 'foo'
@@ -508,6 +544,19 @@ class Store(KvPersister):
     >>> s._data_of_obj=lambda obj: chr(obj)
     >>> s._obj_of_data=lambda data: ord(data)
     >>> test_store(s)
+
+    Note on defining your own "Mapping Views".
+
+    When you do a `.keys()`, a `.values()` or `.items()` you're getting a `MappingView`
+    instance; an iterable and sized container that provides some methods to access
+    particular aspects of the wrapped mapping.
+
+    If you need to customize the behavior of these instances, you should avoid
+    overriding the `keys`, `values` or `items` methods directly, but instead
+    override the `KeysView`, `ValuesView` or `ItemsView` classes that they use.
+
+    For more, see: https://github.com/i2mint/dol/wiki/Mapping-Views
+
     """
 
     _state_attrs = ['store', '_class_wrapper']
@@ -518,28 +567,23 @@ class Store(KvPersister):
 
         if isinstance(store, type):
             store = store()
+
         self.store = store
 
+        if hasattr(self.store, 'KeysView'):
+            self.KeysView = self.store.KeysView
+
+        if hasattr(self.store, 'ValuesView'):
+            self.ValuesView = self.store.ValuesView
+
+        if hasattr(self.store, 'ItemsView'):
+            self.ItemsView = self.store.ItemsView
+
     _id_of_key = static_identity_method
     _key_of_id = static_identity_method
     _data_of_obj = static_identity_method
     _obj_of_data = static_identity_method
 
-    KeysView = BaseKeysView
-    ValuesView = BaseValuesView
-    ItemsView = BaseItemsView
-
-    def keys(self):
-        # each of these methods use the factory method on self,
-        # here that's self.KeysView(), and expect it to take specific arguments.
-        return self.KeysView(self)
-
-    def values(self):
-        return self.ValuesView(self)
-
-    def items(self):
-        return self.ItemsView(self)
-
     _max_repr_size = None
 
     _errors_that_trigger_missing = (
@@ -560,7 +604,9 @@ class Store(KvPersister):
         )  # to forward dir to delegated stream as well
 
     def __hash__(self):
-        return self.store.__hash__()
+        return hash(self.store)
+        # changed from the following (store.__hash__ was None sometimes (so not callable)
+        # return self.store.__hash__()
 
     # Read ####################################################################
 
@@ -587,19 +633,11 @@ class Store(KvPersister):
         yield from (self._key_of_id(k) for k in self.store)
         # return map(self._key_of_id, self.store.__iter__())
 
-    # def items(self) -> ItemIter:
-    #     if hasattr(self.store, 'items'):
-    #         yield from ((self._key_of_id(k), self._obj_of_data(v)) for k, v in self.store.items())
-    #     else:
-    #         yield from ((self._key_of_id(k), self._obj_of_data(self.store[k])) for k in self.store.__iter__())
-
     def __len__(self) -> int:
         return len(self.store)
-        # return self.store.__len__()
 
     def __contains__(self, k) -> bool:
         return self._id_of_key(k) in self.store
-        # return self.store.__contains__(self._id_of_key(k))
 
     def head(self) -> Item:
         k = None
@@ -684,19 +722,21 @@ KvStore = Store  # alias with explict name
 ########################################################################################################################
 # walking in trees
 
+from typing import Callable, KT, VT, Any, TypeVar, Iterator
 
+PT = TypeVar('PT')  # Path Type
 inf = float('infinity')
 
 
-def val_is_mapping(p, k, v):
+def val_is_mapping(p: PT, k: KT, v: VT) -> bool:
     return isinstance(v, Mapping)
 
 
-def asis(p, k, v):
+def asis(p: PT, k: KT, v: VT) -> Any:
     return p, k, v
 
 
-def tuple_keypath_and_val(p, k, v):
+def tuple_keypath_and_val(p: PT, k: KT, v: VT) -> Tuple[PT, VT]:
     if p == ():  # we're just begining (the root),
         p = (k,)  # so begin the path with the first key.
     else:
@@ -704,29 +744,98 @@ def tuple_keypath_and_val(p, k, v):
     return p, v
 
 
-# TODO: More docs and doctests. This one even merits an extensive usage and example tutorial!
+# TODO: More docs and doctests.
+#  This one even merits an extensive usage and example tutorial!
 def kv_walk(
     v: Mapping,
-    yield_func=asis,  # (p, k, v) -> what you want the gen to yield
-    walk_filt=val_is_mapping,  # (p, k, v) -> whether to explore the nested structure v further
-    pkv_to_pv=tuple_keypath_and_val,
-    p=(),
-):
+    yield_func: Callable[[PT, KT, VT], Any] = asis,
+    walk_filt: Callable[[PT, KT, VT], bool] = val_is_mapping,
+    pkv_to_pv: Callable[[PT, KT, VT], Tuple[PT, VT]] = tuple_keypath_and_val,
+    p: PT = (),
+) -> Iterator[Any]:
     """
+    Walks a nested structure of mappings, yielding stuff on the way.
 
-    :param v:
+    :param v: A nested structure of mappings
     :param yield_func: (pp, k, vv) -> what ever you want the gen to yield
     :param walk_filt: (p, k, vv) -> (bool) whether to explore the nested structure v further
     :param pkv_to_pv:  (p, k, v) -> (pp, vv)
         where pp is a form of p + k (update of the path with the new node k)
         and vv is the value that will be used by both walk_filt and yield_func
-    :param p: The path to v
+    :param p: The path to v (used internally, mainly, to keep track of the path)
 
     >>> d = {'a': 1, 'b': {'c': 2, 'd': 3}}
     >>> list(kv_walk(d))
     [(('a',), 'a', 1), (('b', 'c'), 'c', 2), (('b', 'd'), 'd', 3)]
     >>> list(kv_walk(d, lambda p, k, v: '.'.join(p)))
     ['a', 'b.c', 'b.d']
+
+    The `walk_filt` argument allows you to control what values the walk encountered
+    should be walked through. This also means that this function is what controls
+    when to stop the recursive traversal of the tree, and yield an actual "leaf".
+
+    Say we want to get (path, values) items from a nested mapping/store based on
+    a ``levels`` argument that determines what the desired values are.
+    This can be done as follows:
+
+    >>> def mk_level_walk_filt(levels):
+    ...     return lambda p, k, v: len(p) < levels - 1
+    ...
+    >>> def leveled_map_walk(m, levels):
+    ...     yield from kv_walk(
+    ...         m,
+    ...         yield_func=lambda p, k, v: (p, v),
+    ...         walk_filt=mk_level_walk_filt(levels)
+    ...     )
+    >>> m = {
+    ...     'a': {'b': {'c': 42}},
+    ...     'aa': {'bb': {'cc': 'dragon_con'}}
+    ... }
+    >>>
+    >>> assert (
+    ...         list(leveled_map_walk(m, 3))
+    ...         == [
+    ...             (('a', 'b', 'c'), 42),
+    ...             (('aa', 'bb', 'cc'), 'dragon_con')
+    ...         ]
+    ... )
+    >>> assert (
+    ...         list(leveled_map_walk(m, 2))
+    ...         == [
+    ...             (('a', 'b'), {'c': 42}),
+    ...             (('aa', 'bb'), {'cc': 'dragon_con'})
+    ...         ]
+    ... )
+    >>>
+    >>> assert (
+    ...         list(leveled_map_walk(m, 1))
+    ...         == [
+    ...             (('a',), {'b': {'c': 42}}),
+    ...             (('aa',), {'bb': {'cc': 'dragon_con'}})
+    ...         ]
+    ... )
+
+    Tip: If you want to use ``kv_filt`` to search and extract stuff from a nested
+    mapping, you can have your ``yield_func`` return a sentinel (say, ``None``) to
+    indicate that the value should be skipped, and then filter out the ``None``s from
+    your results.
+
+    >>> mm = {
+    ...     'a': {'b': {'c': 42}},
+    ...     'aa': {'bb': {'cc': 'meaning_of_life'}},
+    ...     'aaa': {'bbb': 314},
+    ... }
+    >>> return_path_if_int_leaf = lambda p, k, v: (p, v) if isinstance(v, int) else None
+    >>> list(filter(None, kv_walk(mm, yield_func=return_path_if_int_leaf)))
+    [(('a', 'b', 'c'), 42), (('aaa', 'bbb'), 314)]
+
+    This "path search" functionality is available as a function in the ``recipes``
+    module, as ``search_paths``.
+
+    Inspiration: ``kv_walk`` was inspired by `remap` from the `boltons` package.
+    You may consider using that instead, as it has a much more extensive documetation:
+    See https://sedimental.org/remap.html for example.
+
     """
     # print(f"1: entered with: v={v}, p={p}")
     for k, vv in v.items():
@@ -891,10 +1000,12 @@ class Stream:
     ['1', '2', '3']
 
     Recipes:
-    - _pre_iter: involving itertools.islice to skip header lines
-    - _pre_iter: involving enumerate to get line indices in stream iterator
-    - _pre_iter = functools.partial(map, line_pre_proc_func) to preprocess all lines with line_pre_proc_func
-    - _pre_iter: include filter before obj
+
+    .. hlist::
+        * _pre_iter: involving itertools.islice to skip header lines
+        * _pre_iter: involving enumerate to get line indices in stream iterator
+        * _pre_iter = functools.partial(map, line_pre_proc_func) to preprocess all lines with line_pre_proc_func
+        * _pre_iter: include filter before obj
     """
 
     def __init__(self, stream):
