@@ -6,6 +6,7 @@ import re
 import os
 from functools import partial, wraps
 from types import MethodType
+from typing import Union
 
 from dol.signatures import set_signature_of_func
 from dol.errors import KeyValidationError, _assert_condition
@@ -358,7 +359,7 @@ def mk_prefix_templates_dicts(template):
 
 
 def mk_kwargs_trans(**trans_func_for_key):
-    """ Make a dict transformer from functions that depends solely on keys (of the dict to be transformed)
+    """Make a dict transformer from functions that depends solely on keys (of the dict to be transformed)
     Used to easily make process_kwargs and process_info_dict arguments for LinearNaming.
     """
     assert all(
@@ -398,13 +399,13 @@ def _mk(self, *args, **kwargs):
     return self.template.format(**kwargs)
 
 
-from dol.trans import add_wrapper_method
-
-# @add_wrapper_method
+# from dol.trans import add_wrapper_method
+#
+# # @add_wrapper_method
 class StrTupleDict(object):
     def __init__(
         self,
-        template: (str, tuple, list),
+        template: Union[str, tuple, list],
         format_dict=None,
         process_kwargs=None,
         process_info_dict=None,
