@@ -2,13 +2,15 @@
 import os
 import shutil
 import re
-from collections import namedtuple, defaultdict
+from collections import deque, namedtuple, defaultdict
 from warnings import warn
-from typing import Any, Hashable, Callable, Iterable, Optional, Union
+from typing import Any, Hashable, Callable, Iterable, Optional, Union, Mapping
 from functools import update_wrapper as _update_wrapper
 from functools import wraps as _wraps
 from functools import partialmethod, partial, WRAPPER_ASSIGNMENTS
 from types import MethodType
+from inspect import Signature, signature, Parameter
+
 
 # monkey patching WRAPPER_ASSIGNMENTS to get "proper" wrapping (adding defaults and kwdefaults
 wrapper_assignments = (*WRAPPER_ASSIGNMENTS, '__defaults__', '__kwdefaults__')
@@ -16,6 +18,145 @@ wrapper_assignments = (*WRAPPER_ASSIGNMENTS, '__defaults__', '__kwdefaults__')
 update_wrapper = partial(_update_wrapper, assigned=wrapper_assignments)
 wraps = partial(_wraps, assigned=wrapper_assignments)
 
+exhaust = partial(deque, maxlen=0)
+
+
+def add_as_attribute_of(obj, name=None):
+    """Decorator that adds a function as an attribute of a container object ``obj``.
+
+    If no ``name`` is given, the ``__name__`` of the function will be used, with a
+    leading underscore removed. This is useful for adding helper functions to main
+    "container" functions without polluting the namespace of the module, at least
+    from the point of view of imports and tab completion.
+
+    >>> def foo():
+    ...    pass
+    >>>
+    >>> @add_as_attribute_of(foo)
+    ... def _helper():
+    ...    pass
+    >>> hasattr(foo, 'helper')
+    True
+    >>> callable(foo.helper)
+    True
+
+    In reality, any object that has a ``__name__`` can be added to the attribute of
+    ``obj``, but the intention is to add helper functions to main "container" functions.
+
+    """
+
+    def _decorator(f):
+        attrname = name or f.__name__
+        if attrname.startswith('_'):
+            attrname = attrname[1:]  # remove leading underscore
+        setattr(obj, attrname, f)
+        return f
+
+    return _decorator
+
+
+# TODO: Deprecate? More complete version here: dol.pat.path_get
+#  Could argue to keep chain_get because simple and straightforward
+#  https://github.com/i2mint/dol/blob/7fc78634ef5a6a11ab8417e7fa8a007852699851/dol/paths.py#L118
+def chain_get(d: Mapping, keys, default=None):
+    """
+    Returns the ``d[key]`` value for the first ``key`` in ``keys`` that is in ``d``, and default if none are found
+
+    Note: Think of ``collections.ChainMap`` where you can look for a single key in a sequence of maps until we find it.
+    Here we look for a sequence of keys in a single map, stopping as soon as we find a key that the map has.
+
+    >>> d = {'here': '&', 'there': 'and', 'every': 'where'}
+    >>> chain_get(d, ['not there', 'not there either', 'there', 'every'])
+    'and'
+
+    Notice how ``'not there'`` and ``'not there either'`` are skipped, ``'there'`` is found and used to retrieve
+    the value, and ``'every'`` is not even checked (because ``'there'`` was found).
+    If non of the keys are found, ``None`` is returned by default.
+
+    >>> assert chain_get(d, ('none', 'of', 'these')) is None
+
+    You can change this default though:
+
+    >>> chain_get(d, ('none', 'of', 'these'), default='Not Found')
+    'Not Found'
+
+    """
+    for key in keys:
+        if key in d:
+            return d[key]
+    return default
+
+
+class Literal:
+    """An object to indicate that the value should be considered literally.
+
+    >>> t = Literal(42)
+    >>> t.get_val()
+    42
+    >>> t()
+    42
+
+    """
+
+    def __init__(self, val):
+        self.val = val
+
+    def get_val(self):
+        """Get the value wrapped by Literal instance.
+
+        One might want to use ``literal.get_val()`` instead ``literal()`` to get the
+        value a ``Literal`` is wrapping because ``.get_val`` is more explicit.
+
+        That said, with a bit of hesitation, we allow the ``literal()`` form as well
+        since it is useful in situations where we need to use a callback function to
+        get a value.
+        """
+        return self.val
+
+    __call__ = get_val
+
+
+def _isinstance(obj, class_or_tuple):
+    """The same as the builtin isinstance, but without the position only restriction,
+    allowing us to use partial to define filter functions for specific types
+    """
+    return isinstance(obj, class_or_tuple)
+
+
+def instance_checker(*types):
+    """Makes a filter function that checks the type of an object.
+
+    >>> f = instance_checker(int, float)
+    >>> f(1)
+    True
+    >>> f(1.0)
+    True
+    >>> f('1.0')
+    False
+    """
+    return partial(_isinstance, class_or_tuple=types)
+
+
+def not_a_mac_junk_path(path: str):
+    """A function that will tell you if the path is not a mac junk path/
+    More precisely, doesn't end with '.DS_Store' or have a `__MACOSX` folder somewhere
+    on it's way.
+
+    This is usually meant to be used with `filter` or `filt_iter` to "filter in" only
+    those actually wanted files (not the junk that mac writes to your filesystem).
+
+    These files annoyingly show up often in zip files, and are usually unwanted.
+
+    See https://apple.stackexchange.com/questions/239578/compress-without-ds-store-and-macosx
+
+    >>> paths = ['A/normal/path', 'A/__MACOSX/path', 'path/ending/in/.DS_Store', 'foo/b']
+    >>> list(filter(not_a_mac_junk_path, paths))
+    ['A/normal/path', 'foo/b']
+    """
+    if path.endswith('.DS_Store') or '__MACOSX' in path.split(os.path.sep):
+        return False  # This is indeed math junk (so filter out)
+    return True  # this is not mac junk (you can keep it)
+
 
 def inject_method(obj, method_function, method_name=None):
     """
@@ -42,20 +183,23 @@ def _disabled_clear_method(self):
     """The clear method is disabled to make dangerous difficult.
     You don't want to delete your whole DB
     If you really want to delete all your data, you can do so by doing something like this:
-        ```
+
+    .. code-block:: python
+
         for k in self:
             del self[k]
-        ```
+        
 
     or (in some cases)
 
-        ```
+    .. code-block:: python
+
         for k in self:
             try:
                 del self[k]
             except KeyError:
                 pass
-        ```
+        
     """
     raise NotImplementedError(f'Instance of {type(self)}: {self.clear.__doc__}')
 
@@ -91,6 +235,147 @@ _delete_keys_one_by_one.disabled = False
 _delete_keys_one_by_one_with_keyerror_supressed.disabled = False
 
 
+# Note: Pipe code is completely independent (with inspect imports signature & Signature)
+#  If you only need simple pipelines, use this, or even copy/paste it where needed.
+# TODO: Public interface mis-aligned with i2. funcs list here, in i2 it's dict. Align?
+#  If we do so, it would be a breaking change since any dependents that expect funcs
+#  to be a list of funcs will iterate over a iterable of names instead.
+class Pipe:
+    """Simple function composition. That is, gives you a callable that implements input -> f_1 -> ... -> f_n -> output.
+
+    >>> def foo(a, b=2):
+    ...     return a + b
+    >>> f = Pipe(foo, lambda x: print(f"x: {x}"))
+    >>> f(3)
+    x: 5
+    >>> len(f)
+    2
+
+    You can name functions, but this would just be for documentation purposes.
+    The names are completely ignored.
+
+    >>> g = Pipe(
+    ...     add_numbers = lambda x, y: x + y,
+    ...     multiply_by_2 = lambda x: x * 2,
+    ...     stringify = str
+    ... )
+    >>> g(2, 3)
+    '10'
+    >>> len(g)
+    3
+
+    Notes:
+        - Pipe instances don't have a __name__ etc. So some expectations of normal functions are not met.
+        - Pipe instance are pickalable (as long as the functions that compose them are)
+
+    You can specify a single functions:
+
+    >>> Pipe(lambda x: x + 1)(2)
+    3
+
+    but
+
+    >>> Pipe()
+    Traceback (most recent call last):
+      ...
+    ValueError: You need to specify at least one function!
+
+    You can specify an instance name and/or doc with the special (reserved) argument
+    names ``__name__`` and ``__doc__`` (which therefore can't be used as function names):
+
+    >>> f = Pipe(map, add_it=sum, __name__='map_and_sum', __doc__='Apply func and add')
+    >>> f(lambda x: x * 10, [1, 2, 3])
+    60
+    >>> f.__name__
+    'map_and_sum'
+    >>> f.__doc__
+    'Apply func and add'
+
+    """
+
+    funcs = ()
+
+    def __init__(self, *funcs, **named_funcs):
+        named_funcs = self._process_reserved_names(named_funcs)
+        funcs = list(funcs) + list(named_funcs.values())
+        self.funcs = funcs
+        n_funcs = len(funcs)
+        if n_funcs == 0:
+            raise ValueError('You need to specify at least one function!')
+
+        elif n_funcs == 1:
+            other_funcs = ()
+            first_func = last_func = funcs[0]
+        else:
+            first_func, *other_funcs = funcs
+            *_, last_func = other_funcs
+
+        self.__signature__ = Pipe._signature_from_first_and_last_func(
+            first_func, last_func
+        )
+        self.first_func, self.other_funcs = first_func, other_funcs
+
+    _reserved_names = ('__name__', '__doc__')
+
+    def _process_reserved_names(self, named_funcs):
+        for name in self._reserved_names:
+            if (value := named_funcs.pop(name, None)) is not None:
+                setattr(self, name, value)
+        return named_funcs
+
+    def __call__(self, *args, **kwargs):
+        out = self.first_func(*args, **kwargs)
+        for func in self.other_funcs:
+            out = func(out)
+        return out
+
+    def __len__(self):
+        return len(self.funcs)
+
+    _dflt_signature = Signature.from_callable(lambda *args, **kwargs: None)
+
+    @staticmethod
+    def _signature_from_first_and_last_func(first_func, last_func):
+        try:
+            input_params = signature(first_func).parameters.values()
+        except ValueError:  # function doesn't have a signature, so take default
+            input_params = Pipe._dflt_signature.parameters.values()
+        try:
+            return_annotation = signature(last_func).return_annotation
+        except ValueError:  # function doesn't have a signature, so take default
+            return_annotation = Pipe._dflt_signature.return_annotation
+        return Signature(tuple(input_params), return_annotation=return_annotation)
+
+
+def _flatten_pipe(pipe):
+    for func in pipe.funcs:
+        if isinstance(func, Pipe):
+            yield from _flatten_pipe(func)
+        else:
+            yield func
+
+
+def flatten_pipe(pipe):
+    """
+    Unravel nested Pipes to get a flat 'sequence of functions' version of input.
+
+    >>> def f(x): return x + 1
+    >>> def g(x): return x * 2
+    >>> def h(x): return x - 3
+    >>> a = Pipe(f, g, h)
+    >>> b = Pipe(f, Pipe(g, h))
+    >>> len(a)
+    3
+    >>> len(b)
+    2
+    >>> c = flatten_pipe(b)
+    >>> len(c)
+    3
+    >>> assert a(10) == b(10) == c(10) == 19
+    """
+    return Pipe(*_flatten_pipe(pipe))
+
+
 def partialclass(cls, *args, **kwargs):
     """What partial(cls, *args, **kwargs) does, but returning a class instead of an object.
 
@@ -100,6 +385,7 @@ def partialclass(cls, *args, **kwargs):
     The raison d'être of partialclass is that it returns a type, so let's have a look at that with
     a useless class.
 
+    >>> from inspect import signature
     >>> class A:
     ...     pass
     >>> assert isinstance(A, type) == isinstance(partialclass(A), type) == True
@@ -140,7 +426,7 @@ def partialclass(cls, *args, **kwargs):
     Note in the last partial that since ``a`` was fixed, you need to specify the keyword ``AA(a=3)``.
     ``AA(3)`` won't work:
 
-    >>> AA(3)
+    >>> AA(3)  # doctest: +SKIP
     Traceback (most recent call last):
       ...
     TypeError: __init__() got multiple values for argument 'a'
@@ -152,7 +438,7 @@ def partialclass(cls, *args, **kwargs):
     >>> assert str(signature(AA)) == '(b=1)'
     >>> assert str(AA(3)) == 'A(a=22, b=3)'
 
-    ```
+    
     """
     assert isinstance(cls, type), f'cls should be a type, was a {type(cls)}: {cls}'
 
@@ -234,9 +520,11 @@ def norm_kv_filt(kv_filt: Callable[[Any], bool]):
     Only thing: Your function needs to call a key `k` and a value `v`.
     But hey, it's alright, if you have a function that calls things differently, just do
     something like
-    ```
+
+    .. code-block:: python
+
         new_filt_func = lambda k, v: your_filt_func(..., key=k, ..., value=v, ...)
-    ```
+    
     and all will be fine.
 
     :param kv_filt: callable (starting with signature (k), (v), or (k, v)), and returning  a boolean
@@ -381,7 +669,8 @@ def groupby(
     val: Optional[Callable[[Item], Any]] = None,
     group_factory=list,
 ) -> dict:
-    """Groups items according to group keys updated from those items through the given (item_to_)key function.
+    """Groups items according to group keys updated from those items through the given
+    (item_to_)key function.
 
     Args:
         items: iterable of items
@@ -422,7 +711,7 @@ def groupby(
 
 
 def regroupby(items, *key_funcs, **named_key_funcs):
-    """REcursive groupby. Applies the groupby function recursively, using a sequence of key functions.
+    """Recursive groupby. Applies the groupby function recursively, using a sequence of key functions.
 
     Note: The named_key_funcs argument names don't have any external effect.
         They just give a name to the key function, for code reading clarity purposes.
@@ -463,8 +752,6 @@ GroupReleaseCond = Union[
     Callable[[Groups, GroupKey, GroupItems], bool],
 ]
 
-from inspect import signature
-
 
 def igroupby(
     items: Iterable[Item],
@@ -798,6 +1085,7 @@ class MutableStruct(Struct):
 def max_common_prefix(a):
     """
     Given a list of strings (or other sliceable sequences), returns the longest common prefix
+    
     :param a: list-like of strings
     :return: the smallest common prefix of all strings in a
     """
@@ -891,10 +1179,12 @@ class ImmutableMixin:
     popitem = _immutable
 
 
-class imdict(dict, HashableMixin, ImmutableMixin):
-    """ A frozen hashable dict """
-
-    pass
+# TODO: Lint still considers instances of imdict to be mutable.
+#  Probably because it still sees the mutator methods in the class definition.
+#  Maybe I should just remove them from the class definition?
+# TODO: Generalize to a function that makes any class immutable.
+class imdict(ImmutableMixin, dict, HashableMixin):
+    """A frozen hashable dict"""
 
 
 def move_files_of_folder_to_trash(folder):
@@ -972,5 +1262,38 @@ class ModuleNotFoundIgnore:
         return True
 
 
+def num_of_required_args(func):
+    """Number or REQUIRED arguments of a function.
+
+    Contrast the behavior below with that of ``num_of_args``, which counts all
+    parameters, including the variadics and defaulted ones.
+
+    >>> num_of_required_args(lambda a, b, c: None)
+    3
+    >>> num_of_required_args(lambda a, b, c=3: None)
+    2
+    >>> num_of_required_args(lambda a, *args, b, c=1, d=2, **kwargs: None)
+    2
+    """
+    var_param_kinds = {Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD}
+    sig = signature(func)
+    return sum(
+        1
+        for p in sig.parameters.values()
+        if p.default is Parameter.empty and p.kind not in var_param_kinds
+    )
+
+
 def num_of_args(func):
+    """Number of arguments (parameters) of the function.
+
+    Contrast the behavior below with that of ``num_of_required_args``.
+
+    >>> num_of_args(lambda a, b, c: None)
+    3
+    >>> num_of_args(lambda a, b, c=3: None)
+    3
+    >>> num_of_args(lambda a, *args, b, c=1, d=2, **kwargs: None)
+    6
+    """
     return len(signature(func).parameters)
